const char *program_name = "MacLucasFFTW"; /* for perror() and similar */
#include "version.h"
char version[128]; /* overly long, but certain */

/* The following comment is by Guillermo Ballester Valor.  Feel free to send me email
   about problems as well.  --Will Edgington, wedgingt@acm.org */

/* MacLucasFFTW.c
   
   lucaslfftw v.1.0.0
   Guillermo Ballester Valor, Oct. 1999
   gbv@ctv.es
   
   This is an adaptation of Richard Crandall lucdwt.c and Sweeney 
   MacLucasUNIX.c code.  There are few things mine own.

   The memory requirements is about q bytes, where q is the exponent of
   mersenne number being tested. (m(q))

   This modification is designed to use the GNU package FFTW (the Fastest 
   Fourier Transform in the West). http://www.fftw.org

   I would like to know if this is true. I've already made some adaptation
   of GNU-GMP pakage and really is the fastest of the general C-written FFT 
   packages I tried (in a Pentium 166-mmx). This has some important 
   advantages:
    
   1) It can use other than power of two FFT's lenghts.

   2) For a given FFT length, It can search the best FFT scheme in a system.
      When it had found is 'optimal' solution, it can store it in a file or
      string and use it whenever it needed.

   3) It can be called with MULTITREADS flags to be used in system with more
      than one processor.

      
   At the monment, what I made is :
   a) Remove all FFT code from lucdwt and the awful add_signal and 
   patch routines. It has been replaced by an adaptation of MacLucasUNIX
   normalize, normalize_last and the main routines.
 
   b) Change the code lines needed to support other lengths than power of two.
    
   c) Introduce some small routines to select the FFT lengths.
   
   d) Use the 'fftw.pln' file generated by my own tunefftw program
      for maximum performance with FFTW package.
   
   e) and some other minor changes I don't remember.. :-)

   WHAT YOU NEED TO RUN THIS?

   i) Obviously to install FFTW package.
      See http://www.fftw.org
      Its a very vell documented package. I've installed from source files
      with:
      ./configure
      make
      make install
   ii) Make the tune_fftw program. It's very simple to make. The source code
      is joined with this one. The link option must include -lrfftw -lfftw -lm

   iii) Make this simple program. Again the linker option must have -lm -lrfftw -lfftw
      
   iv) And try it like original MacLucasUNIX.

   Please, let me know any problem.

*/

/* Include Files */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>

// If you have libfftw, but not this #include file, let me know;
// Cygwin has fftw3.h instead but no rfftw.h equivalent that I have found.
// If you don't have libfftw, skip compiling this with 'make NO_FFTW=true'.
#include "fftw.h"
#include "rfftw.h"

#ifndef DO_NOT_USE_LONG_DOUBLE
/* This should always be defined at compile time for all .o's that will link with this file's .o */
# define DO_NOT_USE_LONG_DOUBLE
#endif

/* some include needed for mers package */
#include "setup.h"
#include "balance.h"
#include "rw.h"
#include "sq_herm.h"
#include "init_fft.h"

/* some definitions needed by mers package */
#define MASK (0)
#define SHIFT (0)
#define kErrLimit (0.35)
#define kErrChkFreq (100)
#define kErrChk (1)
#define kLast (2)

/* prototypes of other functions in this file that aren't in an include file */
extern void auto_convolution PROTO((BIG_DOUBLE *x, UL N));
extern void init_lucas PROTO((UL q, UL n));
extern BIG_DOUBLE normalize PROTO((BIG_DOUBLE *x, UL N, UL err_flag));
extern BIG_DOUBLE last_normalize PROTO((BIG_DOUBLE *x, UL N, UL err_flag));
extern BIG_DOUBLE lucas_square PROTO((BIG_DOUBLE *x, UL N, UL iter, UL last, UL error_log));
extern UL power_of_two_length PROTO((UL n));
extern UL choose_length PROTO((UL n, UL N));

/************************ definitions ************************************/

/* IF YOU HAVE LOTS OF REGISTERS IN A RISC PROCESSOR DEFINE REGISTER register */
/*#define REGISTER register */
#define REGISTER /*  */

/* this is to read two_to_phi and two_to_minusphi only every UPDATE times 
   see, normalize and normalize_last routines */

#define UPDATE 64

/* This used to try to align to an even multiple of 16 BIG_DOUBLEs in */
/*  Guillermo's original version, but adjusting pointers from malloc() */
/*  causes core dumps when they're later passed to free().  Is there */
/*  another way to do the same thing?  --wedgingt@acm.org */
#define ALLOC_DOUBLES(n) ((BIG_DOUBLE *)malloc((n)*sizeof(BIG_DOUBLE)))

/* global variables needed */
BIG_DOUBLE     *two_to_phi, *two_to_minusphi;
BIG_DOUBLE     high,low,highinv,lowinv;
BIG_DOUBLE     Gsmall,Gbig,Hsmall,Hbig;
UL             b, c;
rfftw_plan     forw,back;
FILE           *wisdom_file;


/********  The TRICK to round to nearest *******************************/
/* This plays with the internal hardward round to nearest when we add a
small number to a very big number, called bigA number. For intel fpu
this is 3*2^62. Then you have to sustract the same number (named different 
to confuse the clever compilers). This makes rint() very fast.
 
If you have an intel machine and you want to use this, define BE_TRICKY 
I'm thinkinig in make an easy tool to detect this tricky number in every
system-compiler, and print it into a single tricky.h file for compilation 
time.
*/

#define BE_TRICKY

#if defined(__i386__) && defined(BE_TRICKY)
BIG_DOUBLE     bigA,bigB;
#endif


/**************************************************************
 *
 *      FFT and other related Functions
 *
 **************************************************************/

/* rint is not ANSI compatible, so we need a definition for 
 * WIN32 and other platforms with rint.
 * Also we use that to write the trick to rint()
 */


#if defined(__i386__) && defined(BE_TRICKY)
# define RINT(x) (((x) + A ) - B)
#else
# define RINT(x) floor((x) + 0.5);
#endif

#ifdef __STDC__
void square_hermitian(BIG_DOUBLE *x, UL n)
#else
void square_hermitian(x, n)
BIG_DOUBLE *x;
UL          n;
#endif
{
  REGISTER UL  k, half = n>>1;
  REGISTER BIG_DOUBLE c2, d;

  x[0] *= x[0];
  if ((n & 1) == 0) x[half] *= x[half];
  for (k=1;k<half;k++)
    {
      c2 = x[k];
      d = x[n-k];
      x[n-k] = 2.0*c2*d;
      x[k] = (c2+d)*(c2-d);
    }
}

#ifdef __STDC__
static void init_fftw(UL n)
#else
static void init_fftw(n)
UL n
#endif
{
  UL fftw_make_wisdom=0;
  wisdom_file=fopen("fftw.pln","r");
  if (wisdom_file != NULL )
    { 
      if (FFTW_FAILURE == fftw_import_wisdom_from_file (wisdom_file))
	{
	  fftw_make_wisdom=1;
	  fclose(wisdom_file);
	}
      else
	{
	  fclose(wisdom_file);
	  forw=rfftw_create_plan(n,FFTW_FORWARD,FFTW_MEASURE | FFTW_USE_WISDOM);
	  back=rfftw_create_plan(n,FFTW_BACKWARD,FFTW_MEASURE | FFTW_USE_WISDOM);
	}
    }
  else fftw_make_wisdom=1;
    
  if(fftw_make_wisdom == 1)
    {
      wisdom_file=fopen("runtime.pln","r+");
      if (wisdom_file == NULL) wisdom_file=fopen("runtime.pln","a+");
      rewind(wisdom_file);
      fftw_import_wisdom_from_file(wisdom_file);
      printf("\n Trying to find the best FFT for length %ld ...",n);
      fflush(stdout);
      forw=rfftw_create_plan(n, FFTW_FORWARD ,FFTW_MEASURE | FFTW_USE_WISDOM );
      back=rfftw_create_plan(n, FFTW_BACKWARD, FFTW_MEASURE | FFTW_USE_WISDOM);
      rewind(wisdom_file);
      fftw_export_wisdom_to_file(wisdom_file);
      fclose(wisdom_file);
      printf(" Done.\n");
    }  
}


#ifdef __STDC__
void auto_convolution(BIG_DOUBLE *x, UL N)
#else
void auto_convolution( x, N)
BIG_DOUBLE *x;
UL          N;
#endif
{
    BIG_DOUBLE * outx=x+N;
    rfftw_one(forw, x, outx);
    square_hermitian(outx,N);
    rfftw_one(back,outx,x);
}



/****************************************************************************
 *           Lucas Test - specific routines                                 *
 ***************************************************************************/

#ifdef __STDC__
void init_lucas(UL q, UL n)
#else
void init_lucas(q,n)
UL  q,n;
#endif
{
  UL    j,qn,a,i,done;
  BIG_DOUBLE log2 = log(2.0);

  two_to_phi = ALLOC_DOUBLES(n/UPDATE);
  two_to_minusphi = ALLOC_DOUBLES(n/UPDATE);

  low = floor((exp(floor((BIG_DOUBLE)q/n)*log2))+0.5);
  high = low+low;
  lowinv = 1.0/low;
  highinv = 1.0/high;
  b = q % n;
  c = n-b;

  two_to_phi[0] = 1.0;
  two_to_minusphi[0] = 1.0/(BIG_DOUBLE)(n);
  qn = (b*UPDATE)%n;

  for(i=1,j=UPDATE; j<n; j+=UPDATE,i++)
    {
      a = n - qn;
      two_to_phi[i] = exp(a*log2/n);
      two_to_minusphi[i] = 1.0/(two_to_phi[i]*n);
      qn+=b*UPDATE;
      qn%=n;
    }

  Hbig = exp(c*log2/n);
  Gbig = 1/Hbig;
  done = 0;
  j = 0;
  while (!done)
    {
      if (!((j*b) % n >=c || j==0))
	{
	  a = n -((j+1)*b)%n;
	  i = n -(j*b)%n;
	  Hsmall = exp(a*log2/n)/exp(i*log2/n);
	  Gsmall = 1/Hsmall;
	  done = 1;
	}
      j++;
    }
}



#ifdef __STDC__
BIG_DOUBLE normalize(BIG_DOUBLE *x,UL N, UL err_flag )
#else
BIG_DOUBLE normalize(x, N, err_flag)
     BIG_DOUBLE *x;
     UL N;
     UL err_flag;
#endif
{
  REGISTER UL i,j,k,lastloop,bj,offset,newOffset;
  REGISTER UL size0;
  REGISTER BIG_DOUBLE hi=high, hiinv=highinv, lo=low, loinv=lowinv;
  REGISTER BIG_DOUBLE temp0,tempErr;
  REGISTER BIG_DOUBLE maxerr=0.0,err=0.0,ttmpSmall=Gsmall,ttmpBig=Gbig,
    ttmp,ttp,ttmpTemp;
  REGISTER BIG_DOUBLE carry,ttpSmall=Hsmall,ttpBig=Hbig;
#if defined(__i386__) && defined(BE_TRICKY)
  REGISTER BIG_DOUBLE A=bigA,B=bigB;
#endif
  carry = - 2.0; /* this is the -2 of the LL x*x - 2 */

  lastloop = N-UPDATE;

  bj=N;
  size0 = 1;
  offset=0;
  
  for (i=0,j=0; j<N; j+=UPDATE,i++) 
    {
      ttmp = two_to_minusphi[i];
      ttp = two_to_phi[i];
      
      for (k=1; k<UPDATE; k++) 
	{
	  temp0 = x[offset];
	  newOffset = j + k;
	  tempErr = RINT( temp0*ttmp );
	  if (err_flag) 
	    {
	      err = fabs(temp0*ttmp-tempErr);
	      if (err>maxerr) maxerr=err;
	    }
	  temp0 = tempErr + carry;
	  if (size0) 
	    {
	      temp0 *= hiinv;
	      carry = RINT(temp0);
	      bj+=b;
	      ttmp *=ttmpBig;
	      if(bj>=N) bj -= N; 
	      x[offset] = (temp0-carry) * ttp * hi;
	      size0 = (bj>=c);
	      ttp *= ttpBig;
	    }
	  else 
	    {
	      temp0 *= loinv;
	      carry = RINT(temp0);
	      bj+=b;
	      ttmp *=ttmpSmall;
	      if(bj>=N) bj -= N; 
	      x[offset] = (temp0-carry) * ttp * lo;
	      size0 = (bj>=c);
	      ttp *= ttpSmall;
	   }
	  offset=newOffset;
	}
      temp0 = x[offset];
      newOffset = j + UPDATE;
      if (j==lastloop) size0 = 0;
      tempErr = RINT(temp0*ttmp);
      if (err_flag) 
	{
	  err = fabs(temp0*ttmp-tempErr);
	  if (err>maxerr) maxerr=err;
	}
      temp0 = tempErr + carry;
      if (size0) 
	{
	  temp0 *= hiinv;
	  carry = RINT(temp0);
	  bj+=b;
	  ttmp *=ttmpBig;
	  if(bj>=N) bj -= N; 
	  x[offset] = (temp0-carry) * ttp * hi;
	  size0 = (bj>=c);
	  ttp *= ttpBig;
	}
      else 
	{
	  temp0 *= loinv;
	  carry = RINT(temp0);
	  bj+=b;
	  ttmp *=ttmpSmall;
	  if(bj>=N) bj -= N; 
	  x[offset] = (temp0-carry) * ttp * lo;
	  size0 = (bj>=c);
	  ttp *= ttpSmall;
	}
      offset=newOffset; 
    }
  bj = N;
  ttmp = two_to_minusphi[0];
  ttp = two_to_phi[0];
  
  k=0;
  while(carry != 0)
    {
      ttmpTemp = x[k]*ttmp*N;
      size0 = (bj>=c);
      bj += b;
      temp0 = (ttmpTemp + carry);
      if(bj >= N) bj-=N;  
      if (size0)
	{
	  temp0 *= hiinv;
	  ttmp *= ttmpBig;
	  carry = RINT(temp0);
	  x[k] = (temp0 - carry)* ttp * hi;
	  ttp *= ttpBig;
	}
      else
	{
	  temp0 *= loinv;
	  ttmp *= ttmpSmall;
	  carry = RINT(temp0);
	  x[k] = (temp0 -carry)* ttp * lo;
	  ttp *= ttpSmall;
	}
      k++;
    }
  return(maxerr);
}

#ifdef __STDC__
BIG_DOUBLE last_normalize(BIG_DOUBLE *x,UL N,UL err_flag )
#else
BIG_DOUBLE last_normalize(x, N, err_flag)
     BIG_DOUBLE *x;
     UL N;
     UL err_flag;
#endif
{
  REGISTER UL i,j,k,lastloop,bj,offset,newOffset;
  REGISTER UL size0;
  REGISTER BIG_DOUBLE hi=high, hiinv=highinv, lo=low, loinv=lowinv;
  REGISTER BIG_DOUBLE temp0,tempErr;
  REGISTER BIG_DOUBLE maxerr=0.0,err=0.0,ttmpSmall=Gsmall,ttmpBig=Gbig,
    ttmp;
  REGISTER BIG_DOUBLE carry;
#if defined(__i386__) && defined(BE_TRICKY)
  REGISTER BIG_DOUBLE A=bigA,B=bigB;
#endif

  carry = - 2.0; /* this is the -2 of the LL x*x - 2 */
  lastloop = N-UPDATE;

  bj=N;
  size0 = 1;
  offset=0;

  for (j=0,i=0; j<N; j+=UPDATE,i++) 
    {
      ttmp = two_to_minusphi[i];
      
      for (k=1; k<UPDATE; k++) 
	{
	  temp0 = x[offset];
	  newOffset = j + k;
	  tempErr = RINT(temp0*ttmp);
	  if (err_flag) 
	    {
	      err = fabs(temp0*ttmp-tempErr);
	      if (err>maxerr) maxerr=err;
	    }
	  temp0 = tempErr + carry;
	  if (size0) 
	    {
	      temp0 *= hiinv;
	      carry = RINT(temp0);
	      bj+=b;
	      ttmp *=ttmpBig;
	      if(bj>=N) bj -= N; 
	      x[offset] = (temp0-carry) * hi;
	      size0 = (bj>=c);
	    }
	  else 
	    {
	      temp0 *= loinv;
	      carry = RINT(temp0);
	      bj+=b;
	      ttmp *=ttmpSmall;
	      if(bj>=N) bj -= N; 
	      x[offset] = (temp0-carry) * lo;
	      size0 = (bj>=c);
	    }
	  offset=newOffset;
	}
      temp0 = x[offset];
      newOffset = j + UPDATE;
      if (j==lastloop) size0 = 0;
      tempErr = RINT(temp0*ttmp);
      if (err_flag) 
	{
	  err = fabs(temp0*ttmp-tempErr);
	  if (err>maxerr) maxerr=err;
	}
      temp0 = tempErr + carry;
      if (size0) 
	{
	  temp0 *= hiinv;
	  carry = RINT(temp0);
	  bj+=b;
	  ttmp *=ttmpBig;
	  if(bj>=N) bj -= N; 
	  x[offset] = (temp0-carry) * hi;
	  size0 = (bj>=c);
	}
      else 
	{
	  temp0 *= loinv;
	  carry = RINT(temp0);
	  bj+=b;
	  ttmp *=ttmpSmall;
	  if(bj>=N) bj -= N; 
	  x[offset] = (temp0-carry) * lo;
	  size0 = (bj>=c);
	}
      offset=newOffset; 
    }
  bj = N;
  k=0;
  while(carry != 0)
    {
      size0 = (bj>=c);
      bj += b;
      temp0 = (x[k] + carry);
      if(bj >= N) bj-=N;  
      if (size0)
	{
	  temp0 *= hiinv;
	  carry = RINT(temp0);
	  x[k] = (temp0 -carry) * hi;
	}
      else
	{
	  temp0 *= loinv;
	  carry = RINT(temp0);
	  x[k] = (temp0 -carry)* lo;
	}
      k++;
    }
  return(maxerr);
}


#ifdef __STDC__
BIG_DOUBLE lucas_square(BIG_DOUBLE *x, UL N,UL iter, UL last,UL error_log)
#else
BIG_DOUBLE lucas_square(x, N, iter, last, error_log)
     BIG_DOUBLE *x;
     UL N, iter, last, error_log;
#endif
{
	REGISTER BIG_DOUBLE    err;

	auto_convolution(x,N);
	if( iter == last) err=last_normalize(x,N,error_log);
	else err=normalize(x,N,error_log);
	return(err);
}


/* This gives smallest power of two great or equal n */

#ifdef __STDC__
UL power_of_two_length(UL n)
#else
UL power_of_two_length(n)
UL n;
#endif 
{
  UL i = (((UL)1) << 31), k = 32;

  do
   {
    k--;
    i >>= 1;
   } while ((i & n) == 0);
  return k;
}

/* Choose the lenght of FFT , n is the power of two preliminary solution,
N is the minimum required length, the return is the estimation of optimal 
lenght.

The estimation is made very rougly. I suposse a prime k pass cost about
      k*lengthFFT cpu time (in some units)

I try the folowing lenghts
     1)  2^(n-4)*3^2    costs about (3+3+2*(n-4))*9*N   N=2^(n-4)
     2)  2^(n-4)*2*5                (2+5+2*(n-4))*10*N  
     3)  2^(n-4)*11                 (11 +2*(n-4))*11*N
     4)  2^(n-4)*2^2*3              (4+3+2*(n-4))*12*N
     5)  2^(n-4)*13                 (13 +2*(n-4))*13*N
     6)  2^(n-4)*2*7                (2+7+2*(n-4))*14*N
     7)  2^(n-4)*3*5                (3+5+2*(n-4))*15*N
     8)  2^(n-4)*2^4                (8 + 2*(n-4))*16*N
*/
#ifdef __STDC__
UL choose_length(UL n, UL N)
#else
UL choose_length(n, N)
UL n, N;
#endif
{
  UL base=(1<<(n-4)),n1,i,np[7]={8,9,13,7,11,7,6},bestN=1<<n;
  UL best;
  n1=2*(n-4);
  best=(8+n1)*16;
  for (i=15;i>8;i--)
    {
      if (N<=(base*i))
	  {
	    if ( best >= (np[15-i]+n1)*i ) 
	    { 
	      bestN = base*i;
	      best = (np[15-i]+n1)*i;
	    };
	  }
      else break;
    }
  if (bestN < 512)
    return(512);
  return bestN;
}

/**************************************************************
 *
 *      Main Function
 *
 **************************************************************/

#ifdef __STDC__ 
int main(int argc, char *argv[])
#else
int main(argc, argv)
     int argc
char **argv
#endif
{
  UL q = 0L, n,  j = 1L, last = 2L, flag, averbits;
  size_t size,k;
  BIG_DOUBLE err, *x =NULL;
  int restarting = 0;
  char M = '\0';
  FILE *infp = NULL, *outfp = NULL, *dupfp = NULL;
  int primenet = 0; /* flag: interact with PrimeNet?  -N option */
  
  two_to_phi = NULL; 
  two_to_minusphi = NULL;
  
#if defined(__i386__) && defined(BE_TRICKY)
  bigA=(((6.0)*0x2000000L)*0x2000000L)*0x800;
  bigB=bigA;
#endif
  
  averbits = (sizeof(BIG_DOUBLE)*5)/2 - 1;
  
#ifdef macintosh
  argc = ccommand(&argv);
#endif
  
  snprintf(version, sizeof(version)/sizeof(char), "%s %s %s", program_name, VERSION, "Ballester");
  setup();
  while (!shouldTerminate)
    {
      do /* while (restarting) */
	{
	  switch ((restarting != 0) ? 3 :
		  input(argc, argv, &q, &n, &j, &err, &x, last, &M, NULL, NULL,
			NULL, NULL, NULL, MASK, SHIFT,&infp, &outfp, &dupfp, 
			&primenet))
	    {
	    case 0: /* no more input */
	      print_time();
	      return(0);
	      
	    case 1: /*something wrong; error message, if any, already printed */
	    default:
	      printf("caso 1\n");
	      print_time();
	      return(1);
	      
	    case 2: /* continuing work from a partial result */
	      printf("caso 2\n");
	      restarting = 1; 
	      /* not the usual sense of restarting (FFT errors too high) */
	      /* size = n + ((n & MASK) >> SHIFT); */ /* supressed */
	      size = n;
	      break;
	      
	    case 3:
	      n = (q-1)/averbits +1;
	      j = power_of_two_length(n);
	      n = choose_length(j,n);
	 
	      if (x != NULL) free((char *)x);
	      size = n;
	      x = ALLOC_DOUBLES(n+n);
	      
	      for (k=1;k<n;k++) x[k]=0.0;
	      x[0] = 4.0;
	      j = 1;
	      break;
	    }
	  if (two_to_phi != NULL) free((char *)two_to_phi);
	  if (two_to_minusphi != NULL) free((char *)two_to_minusphi);
	  
	  if (!restarting)
	    {
	      if (M != 'U' && M != 'I' && last != q - 2)
		(void)fprintf(stderr, "%s: exponent " PRINTF_FMT_UL 
			      " should not need Lucas-Lehmer testing\n",
			      program_name, q);
	      if ((M != 'U' && M != 'I') || last == q - 2)
		continue;
	    }
	  restarting = 0;
	  init_lucas(q, n);
	  init_fftw(n);
	  if(j==1) x[0]*=two_to_minusphi[0]*n;
	  last = q - 2; /* the last iteration done in the primary loop */
	  for ( ; !shouldTerminate && !restarting && j <= last; j++)
	    {
#if (kErrChkFreq > 1)
	      if ((j % kErrChkFreq) == 1 || j < 1000)
#endif
		flag = kErrChk;
#if (kErrChkFreq > 1)
	      else 
		flag = 0;
#endif
	      err = lucas_square(x, n, j, last, flag);
	      if (chkpnt_iterations != 0 && 
		  j % chkpnt_iterations == 2 && 
		  j < last && 
		  check_point(q, n, j , err, x, MASK, SHIFT, primenet) < 0)
		{
		  print_time();
		  return(errno);
		}
	      if (err > kErrLimit) /* n is not big enough; increase it and start over */
		{
		  averbits--;
		  restarting = 1;
		}
	    }
	  if (restarting) continue;
	  else if (j < last) /* not done, but need to quit, for example, by a SIGTERM signal*/
	    {
	      chkpnt_iterations = 0L;
	      return((check_point(q, n, j, err, x, MASK, SHIFT, primenet) 
		      <= 0) ? errno : 0);
	    }
	} while (restarting);
      printbits(x, q, n, (q > 64L) ? 64L : q, b, c, high, low, MASK, SHIFT,
		version, outfp, dupfp, primenet);
      print_time();
    }
  return(0);
}
