#include "version.h"

/* mmtrial.c - (C) R.W.W. Hooft, 1997. Distribute freely. */
/* Modified by Will Edgington, wedgingt@acm.org, as part of the mers package on behalf of GIMPS */

/* Lower these two numbers if you would want to do smaller MM's and
   raise SIEVE_LIMIT if you want to do larger MM's. */

#define BATCH_SIZE ((unsigned long)500000)
#define SIEVE_LIMIT ((unsigned long)500000)

#define FACTOR_PACKAGE (FACTOR_LIBGMP_INT)
#include "setup.h"
#include "factor.h"
#include <stdio.h>

#ifdef macintosh
# include <console.h>
#endif

const char *program_name = NULL;

#define MAX_FILTER_PRIME (7)
#define FIRST_POST_FILTER_PRIME (11)
#define FILTER_SIZE (3*5*7)
#define FILTER_MULT (8)
#define FILTER_MOD (FILTER_MULT*FILTER_SIZE)
unsigned char filter[FILTER_SIZE]; /* NBBY*FILTER_SIZE bits */

static void mkfilter PROTO((void))
 {
  int k;

  for (k = 0; k < FILTER_SIZE; k++)
    filter[k] = (1 << 7) + (1 << 1); /* all factors are either 1 or 7 mod 8 */
  for (k = 0; k < FILTER_MOD; k += 3)
    filter[k/NBBY] &= (unsigned char)~(1 << (k % NBBY));
  for (k = 0; k < FILTER_MOD; k += 5)
    filter[k/NBBY] &= (unsigned char)~(1 << (k % NBBY));
  for (k = 0; k < FILTER_MOD; k += 7)
    filter[k/NBBY] &= (unsigned char)~(1 << (k % NBBY));
 }

#ifdef __STDC__
static int findk(FACTOR ifilter, FACTOR Td, unsigned long *result)
#else
static int findk(ifilter, Td, result)
FACTOR ifilter, Td;
unsigned long *result;
#endif
 {
  static FACTOR K, Wanted, Temp;
  static int initialized = 0;
  if (result == NULL)
    return(-1);
  if (!initialized)
   {
    mpz_inits(K, Wanted, Temp, NULL);
    initialized = 1;
   }
  mpz_set(Wanted, ifilter);
  mpz_sub_ui(Wanted, Wanted, 1);
  mpz_gcd(Temp, Td, ifilter);
  if (mpz_cmp_ui(Temp, 1) != 0)
    return(-1);
  if (mpz_invert(K, Td, ifilter) == 0)
    return(-1);
  mpz_mul(K, K, Wanted);
  *result = mpz_fdiv_r_ui(NULL, K, mpz_get_ui(ifilter));
  return(1);
 }

char sieve[BATCH_SIZE]; /* NBBY*BATCH_SIZE bits */

#ifdef __STDC__
int main(int argc, char **argv)
#else
int main(argc, argv)
int argc;
char **argv;
#endif
 {
  long count, i;
  UL exponent = 0, k = 1, firstk, inode, nextzero, t120, td120, firstzero, ifilter;
  FACTOR Exponent, Trial, TrialDif, P, Remainder, One, Two, N120, iFilter;
  int verbose = 0;

  mpz_init(Exponent);
  mpz_init(Trial);
  mpz_init(TrialDif);
  mpz_init(P);
  mpz_init(Remainder);
  mpz_init(One);
  mpz_init(Two);
  mpz_init(N120);
  mpz_init(iFilter);

#ifdef macintosh
  argc = ccommand(&argv);
#endif
  if (argv == NULL || argv[0] == NULL)
    program_name = "mmtrial";
  else
    for (program_name = argv[0] + strlen(argv[0]); program_name > argv[0] && *program_name != '/'; )
      --program_name;
  if (*program_name == '/')
    ++program_name;
  setup();
  mkfilter();
  setlinebuf(stdout);

  /* parse the command line */
  if (argc < 2)
   {
    fprintf(stderr, "Usage: %s MM initialk; MM=mersenne exponent exponent\n", argv[0]);
    exit(0);
   }
  while (argv[1][0] == '-' && argv[1][1] == 'v')
   {
    argv++; argc--;
    if (argv[0][2] == '\0' && isascii(argv[1][0]) && isdigit((int)argv[1][0]))
     {
      argv++; argc--;
      verbose = atoi(argv[0]);
      continue;
     }
    if (!isascii(argv[0][2]) || !isdigit((int)argv[0][2]))
      verbose++;
    else
      verbose = atoi(argv[0] + 2);
   }
  if (isascii(argv[1][0]) && isdigit((int)argv[1][0]))
    exponent = (UL)atol(argv[1]);
  if (exponent < 5)
   {
    fprintf(stderr, "Give Mersenne exponent exponent > 4 as argument\n");
    exit(1);
   }
  /* Get initial K if specified */
  if (argc > 2 && isascii(argv[2][0]) && isdigit((int)argv[2][0]))
    k = (UL)atol(argv[2]);
  if (k < 1)
    k = 1;
  else if (k > 1)
    --k;
  fac_set_int(FILTER_MOD, &N120);
  /* Calculate the prime which is the mersenne exponent */
  mpz_ui_pow_ui(P, 2, exponent);
  mpz_sub_ui(P, P, 1);
  /* Difference between successive trials */
  fac_mul_int(P, 2, &TrialDif);
  while (!shouldTerminate) /* Repeat batches of NBBY*BATCH_SIZE until signal'd to stop */
   { /* First value of trial */
    fac_mul_int(TrialDif, k, &Trial);
    fac_add_int(Trial, 1, &Trial);
    /* Filter values */
    t120 = mpz_fdiv_ui(Trial, FILTER_MOD);
    td120 = mpz_fdiv_ui(TrialDif, FILTER_MOD);
    firstk = k;
    bzero(sieve, BATCH_SIZE);
    for (inode = 0; inode < NBBY*BATCH_SIZE; inode++, t120 += td120, t120 %= FILTER_MOD)
      sieve[inode/NBBY] |= (char)(((filter[t120/NBBY] >> (t120 % NBBY)) & 1) << (inode % NBBY));
    k += NBBY*BATCH_SIZE;
    if (verbose > 2)
      printf("Last of batch is k=%ld\n", k - 1);
    ifilter = FIRST_POST_FILTER_PRIME;
    mpz_set_ui(iFilter, ifilter);
    /* Calculate the first trial */
    fac_mul_int(TrialDif, firstk, &Trial);
    fac_add_int(Trial, 1, &Trial);
    /* Now check all numbers in the array for divisibility by ifilter */
    for ( ; ifilter < SIEVE_LIMIT && findk(iFilter, TrialDif, &firstzero) >= 0;
          mpz_nextprime(iFilter, iFilter), ifilter = mpz_get_ui(iFilter))
     { /* Next k for which 2*k*p+1 is divisible by ifilter */
      nextzero = (firstk/ifilter)*ifilter + firstzero;
      while (nextzero < firstk)
        nextzero += ifilter;
      while (nextzero < k) /* Remove nodes that are divisible */
       {
	sieve[(nextzero - firstk)/NBBY] &= (char)~(1 << ((nextzero - firstk) % NBBY));
	nextzero += ifilter;
       } /* End of sieving with ifilter */
     } /* End of sieving process */
    count = 0;
    for (inode = 0; inode < NBBY*BATCH_SIZE; inode++)
      if (sieve[inode/NBBY] & (1 << (inode % NBBY)))
        count++;
    if (verbose > 1)
      printf("Number left after sieving is %ld (out of %lu)\n", count, NBBY*BATCH_SIZE);
    for (inode = 0; !shouldTerminate && inode < NBBY*BATCH_SIZE; inode++) /* Do trials */
     {
      if (sieve[inode/NBBY] & (1 << (inode % NBBY)))
       {
        if (verbose > 3)
          printf("M( M( %ld ) )U: k=%lu done\n", exponent, firstk + inode - 1);
        mpz_mul_ui(Trial, TrialDif, firstk + inode); // TrialDif*(firstk+inode) -> Trial
        mpz_add_ui(Trial, Trial, 1); // increment
        mpz_set_ui(Remainder, 2);
        for (i = 0; i < exponent; i++)
          mpz_powm_ui(Remainder, Remainder, 2, Trial);
        if (mpz_cmp_ui(Remainder, 2) == 0)
         {
          printf("M( M( %ld ) )C: ", exponent);
          mpz_out_str(stdout, 10, Trial);
	  printf(", k=%lu\n", firstk + inode);
         }
        if (verbose > 0)
          printf("M( M( %ld ) )U: k=%lu done\n", exponent, firstk + inode);
       }
     }
    if (verbose == 0)
      printf("M( M( %ld ) )U: k=%lu done\n", exponent, firstk + inode - 1);
   } /* Next batch */
  exit(0);
 }
