#!/usr/bin/perl -Tw
# Release a reservation in the db based on the command line.
#   Mostly for the make that updates the db and, the results directory,
#   and other flat files - that makefile is not in this package.

use strict;
use strict 'vars';

BEGIN { push(@INC, '.'); $ENV{'PATH'} = '/usr/bin:/bin'; }

my ($usage) = "release.pl <exponent> <UUID> <kind>\n"
  . "<exponent> is a non-negative integer indicating the Mersenne number to release.\n"
  . "<UUID> is the UUID or program or script name that made the reservation.\n"
  . "<kind> is a single character, usually a 'mers letter'.\n"
  . "Only exponents that have been reserved should be released.\n"
  . "An exponent of 0 is special and should only be used by mersServer.pl, runSPRP.pl,\n"
  . "  and the Makefile that updates the db and the flat files.\n"
  . "Those UUIDs are 'mersServer', 'runSPRP', and 'make'.\n"
  . "<kind> is '!' for those but does not have to be; only the exponent and UUID matter.\n";

use Mersenne;
use MersDB;

my ($version) = MersInfo::version();
print("$0 version $version\n"),
  exit(0)
  if scalar(grep { m/^--version$/ } @ARGV);

print("$usage"),
  exit(0)
  if scalar(grep { m/^-(h|-help)$/ } @ARGV);

print("$0 version $version\n"),
  exit(0)
  if scalar(grep { m/^--check$/ } @ARGV);

$#ARGV == 2
  or die("exponent, UUID, and kind are required\n$usage");

my ($bigintPattern) = MersInfo::patterns('bigint');
my ($uuidPattern) = MersInfo::patterns('uuid');

$ARGV[0] =~ m,$bigintPattern,
  or die("exponent must be a non-negative integer - '$ARGV[0]' was given\n");
my ($e) = $1;
$ARGV[1] =~ m,$uuidPattern,
  or die("UUID can only contain digits, letters, dash, period, and underscore\n");
my ($uuid) = "$1";
$ARGV[2] =~ m,^([?!A-Za-z])$,
  or die("kind can only be a letter or '!' or '?'\n");
my ($kind) = "$1";

my ($mersDB) = MersDB->new();
my ($dbh) = $mersDB->connect('release');

END
{
  my ($rc) = $dbh ? $dbh->disconnect() : 1;
  $rc or die("disconnect from database failed: " . $dbh->errstr . "\n");
}

my (@r) = $mersDB->update('contact', 'reserve', time());
@r && $r[0] or die("could not update 'contact release', " . $mersDB->dbError() . "\n");

@r = $mersDB->update('contact', $uuid, time());
@r && $r[0] or die("could not update 'contact $uuid', " . $mersDB->dbError() . "\n");

@r = $mersDB->release($e, $uuid, $kind);
print("released $e $uuid $kind\n"),
  exit(0)
  if @r && $r[0];
print($mersDB->dbError());
exit(1);
